;----------------------------------------------------------------------------;
;                                   i2c_func.inc                             ;
;----------------------------------------------------------------------------;
;
;   Implements I2C communication protocol on a 12f683 microcontroller.  The
;   SDA and SCL lines are pins 0 and 1 on port GPIO, respectively.  Contrary
;   to the standard I2C specification, these boards respond to two different
;   7-bit addresses, allowing polling of two sensors per board.  The six MSBs
;   of the 7-bit address form the board address, while the LSB selects the
;   sensor.
;
;   Date:       Author:             Comment:
;   --------------------------------------------------------------------------
;   2006-2-27   Jonathan Karpick    Original version
;   2006-8-1    Barrett Heyneman    Reworked to more closely adhere to I2C
;   2006-8-24   Sal Tujillo         Changed method to check for start
;                                   condition to eliminate false starts.	
;-----------------------------------------------------------------------------

#define I2CPORT GPIO
#define I2CTRIS TRISIO
#define SCL_PIN			1
#define SDA_PIN			0

#define SCL I2CPORT, SCL_PIN
#define SCL_TRIS I2CTRIS, SCL_PIN

#define SDA I2CPORT, SDA_PIN
#define SDA_TRIS I2CTRIS, SDA_PIN

; Masks for checking the address
#define GET_START_ADD		B'00001101'
#define ADDR_BITS_MASK		B'11111100'
#define ADDR_MASK			BOARD_ADDRESS << 2
#define RW_BIT				0
#define SENSOR_NUM_BIT		1

#define POLL_BIT			5	; Bit in the poll count for the LED to follow

; Variables for this module
	cblock
		buffer			; buffer for I2C
		polls			; count of # of times polled (so we can blink the LED
						;	at a reasonable speed)
	endc

;------	ResetWDT ------
; Description:		Resets the WDT and sets it to run for another 1ms
;
; Scope:			Private (macro)
; Arguments:		None
; Returns:			None
; Rgstrs. Changed:	W, WDTCON
;----------------------
ResetWDT macro
	clrwdt
	movlw 0x01		; Prescale = 1:32, enable WDT
	movwf WDTCON	; => 1ms WDT
	endm

;------	EndWDT ------
; Description:		Resets the WDT and disables it
;
; Scope:			Private (macro)
; Arguments:		None
; Returns:			None
; Rgstrs. Changed:	WDTCON
;--------------------
EndWDT macro
	clrwdt
	bcf WDTCON,SWDTEN	; Turn off WDT
	endm

;------	InitI2C ------
; Description:		Initializes the software I2C, by setting the pins used for
;					SCA and SCL to inputs and clearing the I2C buffer.
;
; Scope:			Public (function)
; Arguments:		None
; Returns:			None
; Rgstrs Changed:	I2CTRIS, buffer
;---------------------
InitI2C:
	BANKSEL I2CTRIS
	bsf SCL_TRIS
	bsf SDA_TRIS
	BANKSEL I2CPORT
	bcf	SCL
	bcf	SDA
	clrf buffer
	clrf polls
	return

;------	WaitSCLHigh ------
; Description:		Blocking code which waits for SCL to go high.  Will take
;					at least 2 instruction cycles.
;
; Scope:			Private (macro)
; Arguments:		None
; Returns:			None
; Rgstrs. Changed:	None
;-------------------------
WaitSCLHigh macro
	btfss SCL
	goto $-1
	endm

;------ WaitSCLLow ------
; Description:		Blocking code which waits for SCL to go low.  Will take
;					at least 2 instruction cycles.
;
; Scope:			Private (macro)
; Arguments:		None
; Returns:			None
; Rgstrs. Changed:	None
;------------------------
WaitSCLLow macro
	btfsc SCL
	goto $-1
	endm

;------	WaitSDAHigh ------
; Description:		Blocking code which waits for SDA to go high.  Will take
;					at least 2 instruction cycles.
;
; Scope:			Private (macro)
; Arguments:		None
; Returns:			None
; Rgstrs. Changed:	None
;-------------------------
WaitSDAHigh macro
	btfss SDA
	goto $-1
	endm

;------ WaitSDALow ------
; Description:		Blocking code which waits for SDA to go low.  Will take
;					at least 2 instruction cycles.
;
; Scope:			Private (macro)
; Arguments:		None
; Returns:			None
; Rgstrs. Changed:	None
;------------------------
WaitSDALow macro
	btfsc SDA
	goto $-1
	endm
	
	
;------	StretchClock ------
; Description:		Implements clock stretching; SCL is held low preventing
;					any I2C communication until this PIC is ready.  Assumes
;					that FSR points to I2CTRIS.
;
; Scope:			Private (macro)
; Arguments:		None
; Returns:			None
; Rgstrs. Changed:	I2CTRIS, SCL
;--------------------------
StretchClock macro
	bcf		SCL		; Clear the SCL latch
	bcf		INDF, SCL_PIN	; Set SCL to an output, pulling it low
	endm


;------ PutBit ------
; Description:		Assumes that FSR points to the SDA line's TRIS register
;					and that the SDA line's latch is set low.  Then, setting
;					the SDA line to an output sends a '0', while setting the
;					SDA line to an input sends a '1' (I2C spec requires SDA to
;					be pulled up externally).  Assumes FSR points to I2CTRIS.
;					code blocks while waiting for SCL to pulse.
;
; Scope:			Private (macro)
; Arguments:		bit - the bit number in the I2C buffer to put on the SDA
;						line.
; Returns:			None
; Rgstrs Changed:	None
;--------------------
PutBit macro bit
	btfss buffer,bit
		bcf	INDF, SDA_PIN
	WaitSCLHigh
	WaitSCLLow
	bsf		INDF, SDA_PIN
	endm
	
	
;------	PutBitRelease ------
; Description:		Sets up the first data bit for a transmission after the
;					the clock has been stretched.  Because the master will not
;					be holding the clock line low if it has been stretched,
;					the data must be on the SDA line before the slave releases
;					the clock.
;
; Scope:			Private (macro)
; Arguments:		bit - the bit number in the I2C buffer to put on the SDA
;						line.
; Returns:			None
; Rgstrs. Changed:	None
;---------------------------
PutBitRelease macro bit
	btfss buffer, bit
		bcf	INDF, SDA_PIN
	bsf		INDF, SCL_PIN
	WaitSCLHigh
	WaitSCLLow
	bsf		INDF, SDA_PIN
	endm

;------	PutByte ------
; Description:		Puts the full byte in the I2C buffer onto the SDA line,
;					one bit at a time.  Assumes that the clock was stretched
;					and thus makes sure to release the clock after setting up
;					the first data bit.  This code blocks while waiting for
;					SCL to pulse.  Assumes that FSR points to I2CTRIS.
;					Sets the SDA line's latch low and then writes each bit in
;					order.
;
; Scope:			Private (macro)
; Arguments:		None
; Return:			None
; Rgstrs. Changed:	SDA
;---------------------
PutByte macro
	bcf		SDA
	PutBitRelease  7
	PutBit  6
	PutBit  5
	PutBit  4
	PutBit  3
	PutBit  2
	PutBit  1
	PutBit  0
	endm

;------ GetStop ------
; Description:		Blocking code which looks for a stop bit.  I2C protocol
;					defines a stop bit as SDA transitioning low->high while
;					SCL is high.
;
; Scope:			Private (function)
; Arguments:		None
; Returns:			None
; Rgstrs. Changed:	None
;---------------------
GetStop:
	ResetWDT
	WaitSDALow		; Wait until SDA is low
	ResetWDT
	WaitSCLHigh		; Wait until SCL goes high
	ResetWDT
	WaitSDAHigh		; Wait until SDA goes high
	ResetWDT
	btfss SCL		; check to see of SCL is still high
	goto GetStop	; if not, repeat
	EndWDT
	return

;------	SendACK ------
; Description:		Sends an ACK on the I2C channel.  An ACK is defined as a
;					low on SDA for the 9th data bit.  Assumes that FSR points
;					to I2CTRIS.  Clears the SDA line's latch, then sets SDA to
;					an output to pull down the line.
;
; Scope:			Private (macro)
; Arguments:		None
; Returns:			None
; Rgstrs. Changed:	None
;--------------------
SendACK macro
	bcf		SDA		; Clear SDA's latch
	WaitSCLLow		; Make sure we don't change data line until SCL is low
	bcf		INDF, SDA_PIN	; Pull SDA down
	WaitSCLHigh		; Wait for SCL to be pulsed
	WaitSCLLow
	bsf		INDF, SDA_PIN	; Release SDA
	endm

;------	CheckACK ------
; Description:		Checks to see if the next bit received is an ACK or a NACK
;					and sets ackflag appropriately.  This code is blocking
;					while it waits for SCL to pulse.
;
; Scope:			Private (macro)
; Arguments:		None
; Returns:			None
; Rgstrs. Changed	ackflag
CheckACK macro			
	bcf ackflag
	WaitSCLHigh
	btfss SDA
		bsf ackflag
	WaitSCLLow
	endm

;------	GetBit ------
; Description:		Gets the next bit on the SDA line and places it in the
;					given bit of the I2C buffer.  Assumes that the bit has
;					been cleared before execution.
;
; Scope:			Private (macro)
; Arguments:		bit - the bit to alter in the I2C buffer.
; Returns:			None
; Rgstrs. Changed:	buffer
;--------------------
GetBit macro bit
	WaitSCLLow
	WaitSCLHigh
	btfsc SDA
		bsf buffer, bit
	endm

; Do we need this??
;GetBitEnd macro bit
;	WaitClockHigh
;	btfsc SDA
;	bsf buffer,bit
;	endm

;------ GetByte ------
; Description:		Gets the next 8 bits from the SDA line and puts them into
;					the I2C buffer.
;
; Scope:			Private (macro)
; Arguments:		None
; Returns:			None
; Rgstrs. Changed:	buffer
;---------------------
GetByte macro
	clrf buffer
	GetBit 7
	GetBit 6
	GetBit 5
	GetBit 4
	GetBit 3
	GetBit 2
	GetBit 1
	GetBit 0
	endm
	
	
;------ CheckAddr ------
; Description:		Checks to see if the last byte (assumed to be 7 bit
;					addressing) matches this board's address.
;
; Scope:			Private (macro)
; Arguments:		None.
; Returns:			Z - set if the board address is a match
; Rgstrs. Changed:	STATUS
CheckAddr macro
	movf	buffer, W
	andlw	ADDR_BITS_MASK	; Isolate the 6 board address bits
	xorlw	ADDR_MASK		; Check that those match the board address
	endm
	
;------ GetStart ------
; Description:		Blocking code which looks for a start bit.  I2C protocol
;					defines a start bit as SDA transitioning high->low while
;					SCL is high.
;
; Scope:			Private (macro)
; Arguments:		None
; Returns:			None
; Rgstrs. Changed:
;----------------------
GetStart macro
;WaitBothHigh:
	movf	I2CPORT, W		; Grab the current I/O pins
	addlw	GET_START_ADD	; Add the literal => DC set if both are high
	btfss	STATUS, DC
		goto	$-3			; If not, keep waiting for both to be high
	
;WaitNotBothHigh:
	movf	I2CPORT, W		; Grab the current I/O pins
	addlw	GET_START_ADD	; Add the literal => DC set if both are high
	btfsc	STATUS, DC
		goto	$-3			; If not, keep waiting for one/both to go low
	
;CheckStartCondition:
	addlw	B'00000001'		; If we have a start condition, lower nibble of
							;	W = '1111'.  Adding 1 => DC set
	btfss	STATUS, DC
		goto	$-0xA		; If not, wait for both to go high again, otherwise
							;	we're done.
	endm
	
;------ DebugBuffer ------
; Description: 		Displays the contents of the I2C buffer on the debugging
;					LED using two short pulses to represent 0, one longer pulse
;					to represent 1.  Interrupts must be disabled before calling
;					the function, and reenabled if desired after returning.
;
; Scope:			Private
;-----------------------
#define WAIT_VAL	0x03

	cblock
		WaitCounter
	endc

DisplayBit macro bit
	bsf		T_0
	Wait
	btfss	buffer, bit
		bcf		T_0
	Wait
	bsf		T_0
	Wait
	bcf		T_0
	Wait
	Wait
	endm
	
Wait macro
	movlw	WAIT_VAL
	movwf	WaitCounter
	BANKSEL PIR1
	btfss	PIR1, TMR1IF
		goto $-1
	bcf		PIR1, TMR1IF
	BANKSEL	WaitCounter
	decfsz	WaitCounter, f
		goto $-6
	endm
	
DebugBuffer:
	DisplayBit 7
	DisplayBit 6
	DisplayBit 5
	DisplayBit 4
	DisplayBit 3
	DisplayBit 2
	DisplayBit 1
	DisplayBit 0
	Wait
	return
	
	

;------ I2CListen ------
; Description:		Blocking code which listens for a start bit on the I2C
;					line.  Once a start bit and the subsequent address have
;					been received, the address is checked against the addresses
;					for the board.  If the address matches, and a read was
;					requested, an ACK is sent, followed by the correct data.
;					If the address matches, but a write was requested, a NACK
;					is sent.
;
; Scope:			Public (function)
; Arguments:		None
; Returns:			None
; Rgstrs. Changed:	buffer, STATUS, (see functions/macros used)
;-----------------------
I2CListen:	
	movlw	I2CTRIS		; Point FSR to I2CTRIS for all writes/clock stretching
	movwf	FSR			;	(Do this now so that the ACK can be sent as soon
						;	as possible)

	GetStart			; Wait for the start condition.
	
	GetByte				; Read in the next byte (board address, sensor, r/w status)

	btfss	buffer, RW_BIT	; check the R/W bit
		goto GotWriteCommand
	
	CheckAddr			; See if the board address matched
	btfss	STATUS, Z	
		goto	NoMatch
	;goto	AddressMatch
	
AddressMatch:
	SendACK				; Send an ACK indicating that we can respond
	StretchClock		; Stretch the clock so we stay synchronized
	ResetWDT			; Make sure we don't stretch for too long
	
	incf	polls, F	; Increment our poll count
	btfss	polls, POLL_BIT ; Have the LED follow a single bit (it will flicker
		bcf	T_0				;	every 2^n polls)
	btfsc	polls, POLL_BIT
		bsf	T_0
	
	btfss	buffer, SENSOR_NUM_BIT	; See if this was sensor 0 or 1
		movf	sensorvalue0, W
	btfsc	buffer, SENSOR_NUM_BIT
		movf	sensorvalue1, W
	
	movwf	buffer		; Stick the sensor data in the buffer
	PutByte				; Write the data, releasing the clock at the right
						;	time
	goto	DoneListening

NoMatch:	
GotWriteCommand:
DoneListening:	
	EndWDT				; Stop the WDT
	return
